/* Copyright 2003-2005 Wang, Chun-Pin All rights reserved. */
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <syslog.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <pwd.h>

#include "smbftpd.h"
#include "misc.h"
#include "pathnames.h"
#include "param.h"

extern SESSION gSession;

/* Return share path of szShare
 *
 * Return Values:
 *	Share path or NULL
 */
static char *SzGetSharePath(struct SHARE_SET *pValidShare, char *szShare)
{
	struct SHARE_SET *pSet;
	char *szPath = NULL;

	if ((NULL == pValidShare) || (NULL == szShare)){
		syslog(LOG_ERR, "%s (%d) bad parameter", __FILE__, __LINE__);
		return NULL;
	}

	pSet = pValidShare;
	while (pSet) {
		if (strcmp(szShare, pSet->szShare) == 0) {
			szPath = pSet->szPath;
			break;
		}
		pSet = pSet->pNext;
	}
	return szPath;
}

/* Check whether szCurPath is one of share path root
 *
 * Return Values:
 *	TRUE: Yes
 *	FALSE: No
 */
static BOOL IsShareRoot(struct SHARE_SET *pValidShare, char *szCurPath)
{
	struct SHARE_SET *pSet;
	char *szPath;
	BOOL err = FALSE;

	if (NULL == szCurPath){
		syslog(LOG_ERR, "%s (%d) bad parameter", __FILE__, __LINE__);
		return FALSE;
	}

	pSet = pValidShare;
	while (pSet) {
		szPath = pSet->szPath;
		if (strcmp(szPath, szCurPath) == 0) {
			err = TRUE;
			break;
		}
		pSet = pSet->pNext;
	}
	return err;
}

/* This function will replace the real path name to virtual path name.
 * For example, if share "SHARE" has real path named "/volume1/share1",
 * when user access szPath "/volume1/share1/abc/cde", we will change it
 * to "/SHARE/share1/abc/cde". the BufSize is the buffer size of szPath.
 *
 * Return values:
 *  0: Success
 * -1: Failed
 */
int SmbReplaceSharePath(struct SHARE_SET *pValidShare, char *szPath, int BufSize)
{
	struct SHARE_SET *pSet;
	char *szBuf;
	char szSharePath[PATH_MAX + 1];

	if ( (szPath == NULL) || (BufSize == 0) ) {
		syslog(LOG_ERR, "%s (%d) bad parameter", __FILE__, __LINE__);
		return -1;
	}
	if (gSession.DefaultMode == MODE_NORMAL) {
		return 0;
	}

	if (strcmp(szPath, _PATH_SMB_FTPD_ROOT) == 0) {
		snprintf(szPath, BufSize, "/");
		return 0;
	}

	szBuf = (char *) malloc(BufSize+1);
	if (NULL == szBuf) {
		syslog(LOG_ERR, "%s (%d) Failed to malloc(). errno:%d (%s)", 
			   __FILE__, __LINE__, errno, strerror(errno));
		return -1;
	}

	snprintf(szBuf, BufSize+1, "%s/", szPath);

	pSet = pValidShare;
	while (pSet) {
		snprintf(szSharePath, sizeof(szSharePath), "%s/", pSet->szPath);
		if (strncmp(szSharePath, szBuf, strlen(szSharePath)) == 0) {
			snprintf(szPath, BufSize, "/%s%s",
				pSet->szShare,
				szBuf+strlen(szSharePath)-1);
			if (szPath[strlen(szPath) - 1] == '/') {
				szPath[strlen(szPath) - 1] = '\0';
			}
			break;
		}
		pSet = pSet->pNext;
	}
	free(szBuf);
	return 0;
}

/* Check whether szPath is a valid path.
 * Where ShareFlags is the flag that the share should have.
 * For example, when checking whether under writable share,
 * the ShareFlags should contain ATTRIB_WRITABLE flags.
 *
 * Return Values:
 *	TRUE: yes
 *	FALSE: No
 */
static BOOL IsUnderValidPath(struct SHARE_SET *pValidShare, char *szPath, int ShareFlags)
{
	struct SHARE_SET *pSet;
	BOOL err = FALSE;
	char	szTempSharePath[PATH_MAX+1];
	char	szCurPath[PATH_MAX+1], *szSharePath;
	
	if (!szPath) {
		syslog(LOG_ERR, "%s (%d) bad parameter", __FILE__, __LINE__);
		return FALSE;
	}

	bzero(szCurPath, sizeof(szCurPath));

	snprintf(szCurPath, sizeof(szCurPath), "%s/", szPath);

	// for each share, compare the path with the share path
	for (pSet = pValidShare; pSet; pSet = pSet->pNext) {
		if (ShareFlags && !(pSet->flags & ShareFlags)) {
			continue;
		}
		bzero(szTempSharePath, sizeof(szTempSharePath));

		szSharePath = pSet->szPath;
		if (szSharePath == NULL) {continue;}
		
		// Add a "/" in the end of path, which means the /share would become /share/
		snprintf(szTempSharePath, sizeof(szTempSharePath), "%s/", szSharePath);
        
		// User has access right if the prefix of szCurPath is under valid share.
		if (strncmp(szTempSharePath, szCurPath, strlen(szTempSharePath))==0) {
			err = TRUE;
			break;
		}

	} 
    return err;		
	
}

BOOL CheckShareFlag(struct SHARE_SET *pValidShare, char *szPath, int ShareFlags)
{
	return IsUnderValidPath(pValidShare, szPath ,ShareFlags);
}

/* This is the main path parse function.
 * It gets the real path in the system by given szPath
 * We will chdir into each component to check the path.
 * If Flags & FLAG_CHECK_WRITABLE, than check whether szPath is writable.
 * If Flags & FLAG_NO_FOLLOW_LINK, then do not follow the link.
 * If Flags & FLAG_NO_FOLLOW_LAST_LINK, then we won't check the last
 * component.
 *
 * Return Values:
 *      RealPath or NULL
 */
char *SzGetRealPath(struct SHARE_SET *pValidShare, char *szPath, int Flags)
{
	static char szReturnPath[PATH_MAX+1];
	char	szOriginPWD[PATH_MAX+1], szCurDir[PATH_MAX+1];
	char	szDir[PATH_MAX+1], szTmp[PATH_MAX+1];
	char	*szPtr;
	char	*szToken;
	struct	stat statBuf;
	int err;

	if (!szPath) {
		syslog(LOG_ERR, "%s (%d) bad parameter", __FILE__, __LINE__);
		return NULL;
	}

	if (gSession.DefaultMode == MODE_NORMAL) {
		if (Flags & FLAG_NO_FOLLOW_LAST_LINK) {
			strcpy(szReturnPath, szPath);
		} else {
			realpath(szPath, szReturnPath);
		}
		
		return szReturnPath;
	}

	if ((szPath == NULL) || (szPath[0]== '\0')) {
		return NULL;
	}

	bzero(szTmp, sizeof(szTmp));
	snprintf(szTmp, sizeof(szTmp), "%s", szPath);

REDO:
	bzero((void *)&statBuf, sizeof(statBuf));
	bzero(szReturnPath, sizeof(szReturnPath));
	bzero(szOriginPWD, sizeof(szOriginPWD));
	bzero(szDir, sizeof(szDir));

	getcwd(szOriginPWD, sizeof(szOriginPWD));
	// Check whether we are under valid path now
	if ((strcmp(szOriginPWD, _PATH_SMB_FTPD_ROOT) != 0) && 
		(IsUnderValidPath(pValidShare, szOriginPWD, 0) == FALSE)) {
		chdir(_PATH_SMB_FTPD_ROOT);
		return NULL;
	}


	if (szTmp[0] == '/') {
		snprintf(szDir, sizeof(szDir), "_@SMB_FTPD_ROOT@_%s", szTmp);
		strcpy(szReturnPath, "/");
	}else{
		snprintf(szDir, sizeof(szDir), "%s", szTmp);
		snprintf(szReturnPath, sizeof(szReturnPath), "%s", szTmp);
	}

	// separate the path by '/', and chdir into the path to check
	// whether each component name is in the valid path
	for (szToken = strtok(szDir, "/"); szToken != NULL; szToken = strtok(NULL, "/")) {
		// we are under virtual root
		if (strcmp(szToken, "_@SMB_FTPD_ROOT@_") == 0) {
			snprintf(szReturnPath, sizeof(szReturnPath), "%s", _PATH_SMB_FTPD_ROOT);
			chdir(_PATH_SMB_FTPD_ROOT);
			continue;
		}

		bzero(szCurDir, sizeof(szCurDir));
		getcwd(szCurDir, sizeof(szCurDir));

		//  chdir to / when cd .. if we are under the root of share path
		if ( (strcmp(szToken, "..") == 0) && (IsShareRoot(pValidShare, szCurDir) == TRUE) ) {
			snprintf(szReturnPath, sizeof(szReturnPath), "%s", _PATH_SMB_FTPD_ROOT);
			chdir(_PATH_SMB_FTPD_ROOT);
			continue;
		}

		// When under virtual root
		if (strcmp(szCurDir, _PATH_SMB_FTPD_ROOT) == 0) {
			
			// chdir to virtual root when cd .. or . 
			if ((strcmp(szToken, ".") == 0) || (strcmp(szToken, "..") == 0)) {
				strncpy(szReturnPath, _PATH_SMB_FTPD_ROOT, sizeof(szReturnPath)-1);
				chdir(_PATH_SMB_FTPD_ROOT);
				continue;
			// if next path is not valid share path, return NULL
			} else if ( (szPtr = SzGetSharePath(pValidShare, szToken)) == NULL) {
				chdir(szOriginPWD);
				return NULL;
			} else {
				strncpy(szReturnPath, szPtr, sizeof(szReturnPath)-1);
				chdir(szPtr);
				continue;
			}
		} else {
			// We are under valid share path now.
			if(IsUnderValidPath(pValidShare, szCurDir, 0) == FALSE){
				// Not under valid share path
				chdir(szOriginPWD);
				return NULL;
			}

			err = lstat(szToken, (struct stat *) &statBuf);
			if (err != 0) {
				goto LAST_COMPOENT;
			}

			// Dealing with directory
			if (S_ISDIR(statBuf.st_mode)){
				if (chdir(szToken)) {
					// Failed to chdir (directory not exist or permission denied)
					chdir(szOriginPWD);
					return NULL;
				} else {
					bzero(szCurDir, sizeof(szCurDir));
					getcwd(szCurDir, sizeof(szCurDir));
					if (IsUnderValidPath(pValidShare, szCurDir, 0) == TRUE) {
						strncpy(szReturnPath, szCurDir, sizeof(szReturnPath)-1);
						continue;
					} else {
						chdir(szOriginPWD);
						return NULL;
					}
				}
			} else if ( S_ISLNK(statBuf.st_mode) && 
				    !(Flags & FLAG_NO_FOLLOW_LINK) ) {  // Dealing with link

				if (Flags & FLAG_NO_FOLLOW_LAST_LINK) {
					if (*(szToken+strlen(szToken)+1) == '\0') { // Last component
						goto LAST_COMPOENT;
					}
				}
				err = readlink(szToken, szTmp, sizeof(szTmp) - 1);
				if (err < 0) {
					syslog(LOG_ERR, "%s (%d) Why I can lstat, but can't readlink? errno:%d(%s)",
						__FILE__, __LINE__, errno, strerror(errno));
					return NULL;
				} else {
					szTmp[err] = '\0';
				}

				if (*szTmp != '/'){
					SmbReplaceSharePath(pValidShare, szCurDir, sizeof(szCurDir));
					strcpy(szReturnPath, szCurDir);
					snprintf(szCurDir, sizeof(szCurDir), "%s/%s", szReturnPath, szTmp);

					// Re-compose the path and parse again.
					if ( *(szToken + strlen(szToken) + 1) == '\0') {
						snprintf(szTmp, sizeof(szTmp), "%s", szCurDir);
					} else {
						snprintf(szTmp, sizeof(szTmp), "%s/%s", szCurDir, szToken+strlen(szToken)+1);
					}
				} else {
					// Re-compose the path and parse again.
					if ( *(szToken + strlen(szToken) + 1) != '\0') {
						snprintf(szTmp, sizeof(szTmp), "%s/%s", szCurDir, szToken+strlen(szToken)+1);
					}
				}

				chdir(szOriginPWD);
				goto REDO;

			} else {	// Dealing with file
	LAST_COMPOENT:
				if ((Flags & FLAG_CHECK_WRITABLE) && 
				    IsUnderValidPath(pValidShare, szCurDir, ATTRIB_WRITABLE) == FALSE) {
					chdir(szOriginPWD);
					return NULL;
				} else {
					snprintf(szReturnPath, sizeof(szReturnPath), "%s/%s", 
							 szCurDir, szToken);

					// There should be no next dir since it's a file
					if (strtok(NULL, "/")) {
						chdir(szOriginPWD);
						return NULL;
					}
					chdir(szOriginPWD);
					return szReturnPath;
				}
			} // end of file or directory
		} // end of if under virtual root
	} // end of for
	if (Flags & FLAG_CHECK_WRITABLE) {
		getcwd(szCurDir, sizeof(szCurDir));
		if (IsShareRoot(pValidShare, szCurDir) || 
			(IsUnderValidPath(pValidShare, szCurDir, ATTRIB_WRITABLE) == FALSE)) {
			chdir(szOriginPWD);
			return NULL;
		}else{
			chdir(szOriginPWD);
			return szReturnPath;
		}
	} else {
		chdir(szOriginPWD);
		return szReturnPath;
	}
}

/* Free the SharePath */
void ValidShareFree(struct SHARE_SET **ppValidShare)
{
	struct SHARE_SET *p;

	while (*ppValidShare) {
		p = *ppValidShare;
		*ppValidShare = p->pNext;
		if (p->szShare) {
			free(p->szShare);
		}
		if (p->szPath) {
			free(p->szPath);
		}
		free(p);
	}
	
	return;
}

/* Get shares that is accessable by szUser and put in SHARE_INFO
 *
 * Return Values:
 *	0: Success
 *	-1: Failed
 */
int ValidShareGet(char *szUser, struct SMB_SHARE *pSmbShare, struct SHARE_SET **ppValidShare)
{
	struct SMB_SHARE *pCurShare;
	struct stat statCheck;
	char *szHome = NULL;
	int Writable = 0;
	int err = -1;
	BOOL UserDeny;

	if ((NULL == szUser) || (NULL == pSmbShare)) {
		syslog(LOG_ERR, "%s (%d) bad parameter", __FILE__, __LINE__);
		return -1;
	}

	pCurShare = pSmbShare;
	while (pCurShare) {
		szHome = NULL;
		UserDeny = TRUE;
		Writable = 0;

		if (strcmp(pCurShare->szShare, "homes") == 0) {
			char *szPtr = szRealPath(gSession.pwUser->pw_dir);
			if (!szPtr) {
				continue;
			}
			szHome = strdup(szPtr);
			UserDeny = FALSE;
			Writable = 1;
			goto SKIP;
		}

		if (TRUE == IsUserInList(szUser, pCurShare->szRW)) {
			Writable = 1;
			UserDeny = FALSE;
		}

		if (UserDeny == TRUE) {
			if (TRUE == IsUserInList(szUser, pCurShare->szRO)) {
				UserDeny = FALSE;
			}
		}
SKIP:
		// Shared is accessable for szUser
		if (UserDeny == FALSE) {
			if (szHome || ((!stat(pCurShare->szPath, &statCheck)) &&
				(S_ISDIR(statCheck.st_mode)))) {
				struct SHARE_SET *pSet;

				pSet = calloc(1, sizeof(struct SHARE_SET));
				if (pSet == NULL) {
					syslog(LOG_ERR, "%s (%d) Ran out of memory.", __FILE__, __LINE__);
					goto Error;
				}
				pSet->szShare = szHome?strdup("home"):strdup(pCurShare->szShare);
				pSet->szPath = szHome?szHome:strdup(pCurShare->szPath);
				pSet->flags = pCurShare->flags;
				if (Writable) {
					pSet->flags |= ATTRIB_WRITABLE;
				}
				if (IsUserInList(szUser, pCurShare->szDisableRETR)) {
					pSet->flags |= ATTRIB_NO_DOWNLOAD;
				}
				if (IsUserInList(szUser, pCurShare->szDisableLIST)) {
					pSet->flags |= ATTRIB_NO_LIST;
				}
				if (IsUserInList(szUser, pCurShare->szDisableModify)) {
					pSet->flags |= ATTRIB_NO_MODIFY;
				}
				pSet->pNext = *ppValidShare;
				*ppValidShare = pSet;

			}
		}
		pCurShare = pCurShare->pNext;
	}

	err = 0;
Error:
	if (err) {
		ValidShareFree(ppValidShare);
	}
	return 0;
}

static void SMBShareFreeOne(struct SMB_SHARE *pSmbShare)
{
	if (!pSmbShare) {
		return;
	}
	if (pSmbShare->szShare) {
		free(pSmbShare->szShare);
	}
	if (pSmbShare->szPath) {
		free(pSmbShare->szPath);
	}
	if (pSmbShare->szRW) {
		free(pSmbShare->szRW);
	}
	if (pSmbShare->szRO) {
		free(pSmbShare->szRO);
	}
	if (pSmbShare->szDisableRETR) {
		free(pSmbShare->szDisableRETR);
	}
	if (pSmbShare->szDisableLIST) {
		free(pSmbShare->szDisableLIST);
	}
	free(pSmbShare);
}

/* Free all smbshare inforation. */
void SMBShareFree(struct SMB_SHARE **ppSmbShare)
{
	struct SMB_SHARE *p;

	while (*ppSmbShare) {
		p = *ppSmbShare;
		*ppSmbShare = p->pNext;

		if (p->szShare) {
			free(p->szShare);
		}
		if (p->szPath) {
			free(p->szPath);
		}
		if (p->szRW) {
			free(p->szRW);
		}
		if (p->szRO) {
			free(p->szRO);
		}
		if (p->szDisableRETR) {
			free(p->szDisableRETR);
		}
		if (p->szDisableLIST) {
			free(p->szDisableLIST);
		}
		free(p);
	}
}

/* This function will read the configuration in given Samba
 * config file szfSmbConf.
 * It will allocate memory for pSmbShareInfo to store share
 * information. So SMBShareFree() must be call to free it.
 *
 * Return values:
 *	0: success
 *  -1: faild
 */
int SMBShareEnum(char *szfSmbConf, struct SMB_SHARE **ppSmbShare)
{
	struct SMB_SHARE *pPrevSet, *pCurSet;
	FILE *pFile = NULL;
	char szBuf[LINE_MAX], *szPtr1, *szPtr2;
	int err = -1;

	if (NULL == szfSmbConf) {
		syslog(LOG_ERR, "%s (%d) bad parameter", __FILE__, __LINE__);
		return -1;
	}

	pFile = fopen(szfSmbConf, "r");
	if (NULL == pFile) {
		syslog(LOG_ERR, "%s (%d) Failed to open samba config [%s], errno:%d (%s)", 
			   __FILE__, __LINE__, szfSmbConf, errno, strerror(errno));
		goto Error;
	}

	while (NULL != fgets(szBuf, sizeof(szBuf), pFile)) {
		// Skip spaces
		SzTrimSpace(szBuf);
		szPtr1 = szBuf;

		if ((*szPtr1 == '#') || (*szPtr1 == '\0') || (*szPtr1 == '\n')){
			continue;
		} else if (*szPtr1 == '[') {
			szPtr1++;
			szPtr2 = strchr(szPtr1, ']');
			if ((szPtr2 != NULL) && (szPtr2 - szPtr1 > 0)) {
				*szPtr2 = '\0';
				SzTrimSpace(szPtr1);				
				if (strcmp(szPtr1, "global") != 0) {
					struct SMB_SHARE *p;

					p = calloc(1, sizeof(struct SMB_SHARE));
					if (p == NULL) {
						syslog(LOG_ERR, "%s (%d) Out of memory", __FILE__, __LINE__);
						goto Error;
					}
					p->szShare = strdup(szPtr1);
					if (!p->szShare) {
						syslog(LOG_ERR, "%s (%d) Out of memory", __FILE__, __LINE__);
						goto Error;
					}
					p->pNext = *ppSmbShare;
					*ppSmbShare = p;
				}
			}
			
		} else {
			if (!*ppSmbShare) {
				/* We have not gotten a section */
				continue;
			}
			szPtr2 = strchr(szPtr1, '=');
			if (NULL == szPtr2) {
				continue;
			}
			*szPtr2 = '\0';
			szPtr2++;
			SzTrimString(szPtr1);
			SzTrimString(szPtr2);
			if (strcmp(szPtr1, "rw") == 0) {
			   (*ppSmbShare)->szRW = strdup(szPtr2);
				if (!(*ppSmbShare)->szRW) {
					syslog(LOG_ERR, "%s (%d) Out of memory", __FILE__, __LINE__);
					goto Error;
				}
			} else if (strcmp(szPtr1, "ro") == 0) {
				(*ppSmbShare)->szRO = strdup(szPtr2);
				if (!(*ppSmbShare)->szRO) {
					syslog(LOG_ERR, "%s (%d) Out of memory", __FILE__, __LINE__);
					goto Error;
				}
			} else if (strcmp(szPtr1, "disable_download") == 0) {
				(*ppSmbShare)->szDisableRETR = strdup(szPtr2);
				if (!(*ppSmbShare)->szDisableRETR) {
					syslog(LOG_ERR, "%s (%d) Out of memory", __FILE__, __LINE__);
					goto Error;
				}
			} else if (strcmp(szPtr1, "disable_ls") == 0) {
				(*ppSmbShare)->szDisableLIST = strdup(szPtr2);
				if (!(*ppSmbShare)->szDisableLIST) {
					syslog(LOG_ERR, "%s (%d) Out of memory", __FILE__, __LINE__);
					goto Error;
				}
			} else if (strcmp(szPtr1, "disable_modify") == 0) {
				(*ppSmbShare)->szDisableModify = strdup(szPtr2);
				if (!(*ppSmbShare)->szDisableModify) {
					syslog(LOG_ERR, "%s (%d) Out of memory", __FILE__, __LINE__);
					goto Error;
				}
			} else if (strcmp(szPtr1, "path") == 0) {
				char *szPtr = szRealPath(szPtr2);
				if (!szPtr) {
					syslog(LOG_ERR, "%s (%d) path \"%s\" does not exist.", 
						   __FILE__, __LINE__, szPtr2);
				} else {
					(*ppSmbShare)->szPath = strdup(szPtr);
					if (!(*ppSmbShare)->szPath) {
						syslog(LOG_ERR, "%s (%d) Out of memory", __FILE__, __LINE__);
						goto Error;
					}
				}
			} else if (strcmp(szPtr1, "browseable") == 0) {
				if (strcasecmp(szPtr2, "no") == 0) {
					(*ppSmbShare)->flags |= ATTRIB_NO_BROWSE;
				}
			}
		}
	}

	// Remove shares that have no path beside homes
	pPrevSet = pCurSet = *ppSmbShare;
	while (pCurSet) {
		if ((strcmp(pCurSet->szShare, "homes") != 0) && !pCurSet->szPath) {
			if (pPrevSet == pCurSet) { /* Head */
				*ppSmbShare = pCurSet->pNext;
				SMBShareFreeOne(pCurSet);
				pPrevSet = pCurSet = *ppSmbShare;
			} else {
				pPrevSet->pNext = pCurSet->pNext;
				SMBShareFreeOne(pCurSet);
				pCurSet = pPrevSet->pNext;
			}
		} else {
			pPrevSet = pCurSet;
			pCurSet = pCurSet->pNext;
		}
	}
	
	err = 0;
Error:
	if (pFile) {
		fclose(pFile);
	}
	if (err) {
		if (*ppSmbShare != NULL) {
			SMBShareFree(ppSmbShare);
		}
		return -1;
	} else {
		return 0;
	}
}

